// author: Chuck Baker
// CLEO Version 10.0

var EXPORTED_SYMBOLS = [ "CLEO" ];
if (typeof(CLEO) == 'undefined') var CLEO = {};

var Cc = Components.classes;
var Ci = Components.interfaces;
var Cu = Components.utils;
Cu.import("resource://gre/modules/AddonManager.jsm");
Cu.import("resource://gre/modules/Services.jsm");

window.addEventListener('load',function(){CLEO.cleoInit()},true);

//let consoleJSM = Cu.import("resource://gre/modules/Console.jsm", {});
//let CLEOconsole = consoleJSM.console;

// Replacement for deprecated Application.storage
var CLEOstorage = {
	topWin: Cc["@mozilla.org/appshell/window-mediator;1"].getService(Ci.nsIWindowMediator).getMostRecentWindow("navigator:browser"),  
	setItem: function setItem(key, value){
			if(typeof this.topWin.storage == "undefined") this.topWin.storage = {};
			if(key == null || key == "") return false;
			this.topWin.storage[key] = value;
			return true;
	},
	getItem: function getItem(key,dFault){
			if(key == null || key == "") return false;
			if(typeof this.topWin.storage[key] == "undefined") return dFault;
			let value = this.topWin.storage[key];
			return value;
	}
}


CLEO = {
	cleoDesc: [],														// Lines of package description
	cleoExtBUdir: new String(),							// Directory where the results are stored
	cleoExtName: new String(),							// Extension name
	cleoFileList: new Array(),							// Array holding names of items to process
	cleoFilesToAnalizeDir: new String(),		// Path to "filesToAnalize" in temp dir
	cleoFilesToAnalizeDirPath: new String(),
	cleoFilesToPackageDir: new String(),		// Path to "filesToPackage" in temp dir
	cleoHideHelp: Boolean(false),						// Hide help icons?
	cleoInfoFileFlag: Boolean(false),				// Processing view info function?
	cleoInstallDir: new String(),						// Folder where CLEO is installed
	cleoItems: new Number,									// Number of items tp process
	cleoMainWindow: new Object(),						// Main browser window
	cleoMinFEBEver: "10.0",									// Minimum FEBE version required for this CLEO version
	cleoMsg: new Array(),										// Array holding localized strings
	cleoOverwriteOpt: new Number,						// Overwrite package option
	cleoPackageDesc: new String(),					// Package description
	cleoPackageName: new String(),					// Name of finished package
	cleoPathList: new Array(),							// Array holding paths to items to process
	cleoRewriteInfoFlag: Boolean(false),		// Rebuild the comment file?
	cleoTmpDir: new String(),								// nsI pointer to [temp]/febe.tmp/cleo.tmp
	cleoTmpDirName: new String(),						// Name of Firefox temp directory
	cleoVersion: new String(),							// Version of this extension
	cleoViewPackageName: new String(),			// Name of package to view info
	cleoWin: new Object(),									// current Cleo window (Options, etc.)
	febeInstalled: Boolean(false),					// Is FEBE installed?
	
	CLEO_GUID: "CLEO@guid.customsoftwareconsult.com",  			// GUID of this extension
	cFEBE_GUID: "{4BBDD651-70CF-4821-84F8-2B918CF89CA3}",  	// FEBE's GUID
	cleoPrefs: Cc["@mozilla.org/preferences-service;1"].getService(Ci.nsIPrefService).getBranch(""),

cleoLocalize: function (){
	// Error messages, prompts, etc.
	let wm = Cc["@mozilla.org/appshell/window-mediator;1"].getService(Ci.nsIWindowMediator);  
	CLEO.cleoMainWindow = wm.getMostRecentWindow("navigator:browser"); 

	if(!this.cleoMsg[0]){
		let bundle = Services.strings.createBundle("chrome://CLEO/locale/cleo.properties");
		let enumerator = bundle.getSimpleEnumeration();
		while (enumerator.hasMoreElements()) {
			let string = enumerator.getNext().QueryInterface(Ci.nsIPropertyElement);
			let stringName = string.key;
			if(stringName.substr(0,3) != "msg") continue;
			let index = stringName.substr(3);
			index = parseInt(index);
			let value = string.value;
			this.cleoMsg[index] = value;
		}
	}	

	return true;
},
	
getCleoVersion: function (){
	AddonManager.getAddonByID(CLEO.CLEO_GUID, function(addon){
		CLEOstorage.setItem("cleoVersion", addon.version);
		CLEO.cleoVersion = addon.version;
		return true;
	});
},

getFebeVersion: function (){
	try{
		AddonManager.getAddonByID(CLEO.cFEBE_GUID, function(addon){
			let obj = new Object();
			obj.version = addon.version;
			obj.userDisabled = addon.userDisabled;
			obj.installed = true;
			let jsonStr = JSON.stringify(obj);
			CLEOstorage.setItem("cleoFebeVersion", jsonStr);
		});
	}catch(err){
		let obj = new Object();
		obj.version = null;
		obj.userDisabled = null;
		obj.installed = false;
		let jsonStr = JSON.stringify(obj);
		CLEOstorage.setItem("cleoFebeVersion", jsonStr);
	}
	return true;
},

setCleoVersion: function (){
	CLEO.cleoVersion = CLEOstorage.getItem("cleoVersion", "???");
	document.getElementById("cleoAboutVersionID").value = CLEO.cleoVersion;
	return true;
},

cleoAlert: function (message){
	CLEO.cleoWin = window.openDialog("chrome://cleo/content/cleoAlert.xul", "CLEO Alert", "chrome,modal,dialog=yes,alwaysRaised,resizable,titlebar=no,popup=yes",message);	
	CLEO.cleoWin.focus();
	return true;
},

cleoHelp: function (helpIndex,fieldName){
	CLEO.cleoWin = window.openDialog("chrome://cleo/content/cleoHelp.xul", "CLEO Help", "chrome,dialog=yes,alwaysRaised,resizable,titlebar=no,popup=yes",helpIndex,fieldName);
	CLEO.cleoWin.focus();
	return true;
},

cleoTmpName: function (){
  // Return a random filename
  return "CLEO"+Math.floor(Math.random() * 1e+8)+".xpi";
},

cleoOptions: function (aPaneID){
	window.openDialog('chrome://cleo/content/cleoOptions.xul', 'CLEO Options', 'chrome,titlebar,toolbar,centerscreen,resizable');	
},

cleoViewInfo: function (){
	window.openDialog('chrome://cleo/content/cleoViewInfo.xul', 'CLEO ViewInfo', 'chrome,centerscreen,resizable,dependent');	
},

cleoMakePath: function (dir, file){
  // Return the path of file appended to dir
	var aFile = Cc["@mozilla.org/file/local;1"].createInstance(Ci.nsILocalFile);
	aFile.initWithPath(dir);
	aFile.append(file);
	return aFile.path;
},

cleoInit: function (){
	CLEO.getCleoVersion();
	CLEO.getFebeVersion();
	let windowtype = document.documentElement.getAttribute("windowtype");
	if(windowtype == "cleo:FEBEcheck") return true;
	setTimeout(CLEO.checkForFEBE,1000);
	CLEO.cleoLocalize();

	// Get overwrite option
	var prefName = "extensions.cleo.overwriteopt";
	CLEO.cleoOverwriteOpt = CLEO.cleoPrefs.getCharPref(prefName);

	// Get hide help option
	var prefName = "extensions.cleo.hidehelp";
	CLEO.cleoHideHelp = CLEO.cleoPrefs.getBoolPref(prefName);
 
	CLEO.cleoHideHelpIcons();
	window.removeEventListener("load",CLEO.cleoInit,true);
	return true;
},

checkForFEBE: function(){
	if(!CLEO.cleoPrefs.getBoolPref("extensions.cleo.checkforFEBE")) return true;
	let jsonStr = CLEOstorage.getItem("cleoFebeVersion", false);
	let obj = JSON.parse(jsonStr);
	let febeVersion = obj.version
	let isDisabled = obj.userDisabled;
	let installedVersion = obj.version;
	let versionOK = (installedVersion >= CLEO.cleoMinFEBEver);
	let febeInstalled = obj.installed;
	let OK = (febeInstalled && !isDisabled && versionOK);	
	if(!OK) CLEO.openNoFEBEwin();
	//CLEO.openNoFEBEwin();
	return true;
},

openNoFEBEwin: function(){
	CLEO.cleoWin = window.open("chrome://cleo/content/cleoNoFEBE.xul", "CLEO No FEBE", 'chrome,alwaysRaised,centerscreen,resizable');
},

FEBEnotInstalledLoad: function(){
	CLEO.cleoLocalize();
	var msgbox = document.getElementById("cleoNoFebeBoxID");
	var text = document.createElement('text');
	var description = document.createElement('description');
	description.setAttribute('flex', "0");
	var separator = document.createElement('separator');
	separator.setAttribute('class', "thin");

	var textBox = text.cloneNode(false);
	CLEO.cleoVersion = CLEOstorage.getItem("cleoVersion", "???");
	let jsonStr = CLEOstorage.getItem("cleoFebeVersion", false);
	let obj = JSON.parse(jsonStr);
	let febeVersion = obj.version;
	let isDisabled = Boolean(obj.userDisabled);
	let installedVersion = obj.version;
	let versionOK = Boolean(installedVersion >= CLEO.cleoMinFEBEver);
	let febeInstalled = Boolean(obj.installed);
	var type = 0;
	
	if((!febeInstalled) || (febeInstalled && !versionOK) || (febeInstalled && isDisabled)){ 	
		// FEBE is not installed, or is installed but the wrong version, or is installed but disabled
		var txt = CLEO.cleoMsg[00].replace('%CVER%',CLEO.cleoVersion);
		txt = txt.replace('%FVER%',CLEO.cleoMinFEBEver);
		var textBox = text.cloneNode(false);
		textBox.setAttribute('value', txt);
		var desc = description.cloneNode(true);
		desc.appendChild(textBox);
		desc.appendChild(separator);
		msgbox.appendChild(desc);
		type = 1;
	}

	if(febeInstalled && versionOK && isDisabled){ 	
		// FEBE is installed with the correct version, but is disabled
		var txt = CLEO.cleoMsg[07].replace('%FVER%',febeVersion);
		var textBox = text.cloneNode(false);
		textBox.setAttribute('value', txt);
		var desc = description.cloneNode(true);
		desc.appendChild(textBox);
		desc.appendChild(separator);
		msgbox.appendChild(desc);
		type = 2;
	}
	
	if(febeInstalled && !versionOK){	
		// FEBE is installed, but the wrong version
		var txt = CLEO.cleoMsg[03].replace('%FVER%',febeVersion);
		var textBox = text.cloneNode(false);
		textBox.setAttribute('value', txt);
		var desc = description.cloneNode(true);
		desc.appendChild(textBox);
		desc.appendChild(separator);
		msgbox.appendChild(desc);
		type = 3;
	}
	
	function debug(type){
		let br = "\n";
		let txt = "Debug information:"+br;
		if(type == 0) txt += "All tests passed.  You have the correct versions of CLEO and FEBE installed"+br;
		if(type == 1) txt += "FEBE is not installed, or is installed but the wrong version, or is installed but disabled"+br;
		if(type == 2) txt += "FEBE is installed with the correct version, but is disabled"+br;
		if(type == 3) txt += "FEBE is installed, but the wrong version"+br;
		txt += br;
		txt += "FEBE installed: "+Boolean(obj.installed)+br;
		txt += "FEBE version: "+obj.version+br;
		txt += "FEBE disabled: "+Boolean(obj.userDisabled)+br;
		txt += "CLEO min FEBE version: "+CLEO.cleoMinFEBEver+br;
		txt += "CLEO current version: "+CLEO.cleoVersion+br;
		txt += "OK: "+Boolean(installedVersion >= CLEO.cleoMinFEBEver)+br;
		let textBox = document.getElementById('debug');
		let content = document.createTextNode(txt);
		textBox.appendChild(content);
		msgbox.appendChild(textBox);
	}
	
	debug(type);
	
	// Show link text
	var txt = CLEO.cleoMsg[01];
	var textBox = text.cloneNode(false);
	textBox.setAttribute('value', txt);
	var desc = description.cloneNode(true);
	desc.appendChild(textBox);
	desc.appendChild(separator);
	msgbox.appendChild(desc);

	var txt = CLEO.cleoMsg[02];
	var textBox = text.cloneNode(false);
	textBox.setAttribute('value', txt);
	var desc = description.cloneNode(true);
	desc.appendChild(textBox);
	desc.appendChild(separator);
	msgbox.appendChild(desc);
	
	var desc = description.cloneNode(true);
	desc.appendChild(document.createTextNode(''));
	var textBox = text.cloneNode(false);

	textBox.setAttribute('value', "FEBE homepage");
	textBox.setAttribute('class', "link");
	textBox.onclick = function(event){"CLEO.cleoOpenLink('http://softwarebychuck.com/febe/');"};

	desc.appendChild(textBox);
	msgbox.appendChild(desc);

	CLEO.cleoMainWindow.document.getElementById("cleoTLBRbutton").disabled = true;
	return true;
},

cleoGetUnicharPref: function (prefName) {
    return CLEO.cleoPrefs.getComplexValue(prefName, Ci.nsISupportsString).data;
},

cleoSetUnicharPref: function (prefName, prefValue) {
    var supportsString = Cc["@mozilla.org/supports-string;1"].createInstance(Ci.nsISupportsString);
    supportsString.data = prefValue;
    CLEO.cleoPrefs.setComplexValue(prefName, Ci.nsISupportsString, supportsString);

}, 

cleoWindow: function (){
	CLEO.cleoWin=openDialog('chrome://cleo/content/cleo.xul', 'CLEO', 'chrome,centerscreen,resizable');
	return true;
},

cleoSelDir: function (){
  // Select backup destinaton directory
	const nsIFilePicker = Ci.nsIFilePicker;
	var fp = Cc["@mozilla.org/filepicker;1"]
					.createInstance(nsIFilePicker);
	fp.init(window, CLEO.cleoMsg[18], nsIFilePicker.modeGetFolder);
	fp.appendFilters(nsIFilePicker.filterAll | nsIFilePicker.filterText);

  // Set default directory to current backup directory
	if(CLEO.cleoPrefs.prefHasUserValue("extensions.cleo.extBUdir")){
		CLEO.cleoExtBUdir = CLEO.cleoGetUnicharPref("extensions.cleo.extBUdir");
		var dd = Cc["@mozilla.org/file/local;1"]
                     .createInstance(Ci.nsILocalFile);
		if(CLEO.cleoExtBUdir != ""){dd.initWithPath(CLEO.cleoExtBUdir);}
		fp.displayDirectory = dd;
	}
	var rv = fp.show();
	if (rv == nsIFilePicker.returnOK){
		rv = fp.file;
		CLEO.cleoExtBUdir = rv.path;
		CLEO.cleoSetUnicharPref("extensions.cleo.extBUdir", CLEO.cleoExtBUdir);
		document.getElementById("cleoDestDirID").value = CLEO.cleoExtBUdir;
	}
	return true;
},

cleoSetStage: function (){
  // Set up the staging area (cleo.tmp)
	var prefName = "extensions.cleo.extBUdir";
  CLEO.cleoExtBUdir = CLEO.cleoGetUnicharPref(prefName);

	CLEO.cleoTmpDir = Cc["@mozilla.org/file/directory_service;1"]
							 .getService(Ci.nsIProperties)
							 .get("TmpD", Ci.nsIFile);
	CLEO.cleoTmpDir.append("cleo.tmp");

	if(!CLEO.cleoTmpDir.exists() || !CLEO.cleoTmpDir.isDirectory()){   // if it doesn't exist, create
		CLEO.cleoTmpDir.create(Ci.nsIFile.DIRECTORY_TYPE, 0777);
	}
	CLEO.cleoTmpDirName = CLEO.cleoTmpDir.path
	
	// Create "filesToAnalize" directory
	CLEO.cleoFilesToAnalizeDir = CLEO.cleoTmpDir.clone();
	CLEO.cleoFilesToAnalizeDir.append("filesToAnalize");
	CLEO.cleoFilesToAnalizeDir.createUnique(Ci.nsIFile.DIRECTORY_TYPE, 0777);
	
	// Create "filesToPackage" directory
	CLEO.cleoFilesToPackageDir = CLEO.cleoTmpDir.clone();
	CLEO.cleoFilesToPackageDir.append("filesToPackage");
	CLEO.cleoFilesToPackageDir.createUnique(Ci.nsIFile.DIRECTORY_TYPE, 0777);
	FEBE.febeSetMsgs();
	return true;
},

cleoDoPackage: function (){
	CLEO.cleoSetStage();
	CLEO.hideProgressBar(false);
	var packageName = CLEO.cleoPackageName;
	var finalPackageName = packageName + ".xpi";
	
	CLEO.cleoPackageName = CLEO.cleoMakePath(CLEO.cleoExtBUdir, finalPackageName);
	
	var tmpDirDelete = CLEO.cleoTmpDir.clone();
	CLEO.numCleoItems = CLEO.cleoFileList.length

  // Copy the items to the tmp directory
	for (var pos = 0; (pos < CLEO.numCleoItems); pos++){
		CLEO.updateProgress(pos+1, CLEO.numCleoItems, CLEO.cleoFileList[pos]);
		FEBE.febeCopyFile(CLEO.cleoPathList[pos],CLEO.cleoFilesToPackageDir.path,CLEO.cleoFileList[pos]);
	}

  // Copy install.rdf
	CLEO.cleoInstallDir = Cc["@mozilla.org/file/directory_service;1"]
									.getService(Ci.nsIProperties)
									.get("ProfD", Ci.nsIFile);
	CLEO.cleoInstallDir.append("extensions");
	CLEO.cleoInstallDir.append(CLEO.CLEO_GUID);
  CLEO.cleoInstallDir.append("CLEOinstall.rdf");
  FEBE.febeCopyFile(CLEO.cleoInstallDir.path,CLEO.cleoFilesToPackageDir.path,"install.rdf");
     
  // Create description file ("CLEO.cif")
	var file = CLEO.cleoFilesToPackageDir.clone();
	file.append("CLEO.cif");
	file.create(Ci.nsIFile.NORMAL_FILE_TYPE, 0755);
	var cleoDescFile = file.clone();
	
	// Open it
	var descFile = Cc["@mozilla.org/network/file-output-stream;1"]
								.createInstance(Ci.nsIFileOutputStream);
	descFile.init(cleoDescFile, 0x02 | 0x08 | 0x20, 0755, 0); // write, create, truncate
	
	// Setup an output converter stream
	var charset = "UTF-8";
	var osDescFile = Cc["@mozilla.org/intl/converter-output-stream;1"]
                   .createInstance(Ci.nsIConverterOutputStream);
	osDescFile.init(descFile, charset, 4096, 0x0000);
	
	// Write it and close it
	var tmp = document.getElementById("cleoDescription").value
	CLEO.cleoDesc = tmp.split("\n");
	osDescFile.writeString(CLEO.cleoDesc.join("\n"));
	osDescFile.close();	
	var tmpName = CLEO.cleoTmpName();

	if(parseInt(CLEO.cleoOverwriteOpt) == 2){	// Append
		var src = CLEO.cleoMakePath(CLEO.cleoExtBUdir, CLEO.finalPackageName);
		var aFile = Cc["@mozilla.org/file/local;1"]
								.createInstance(Ci.nsILocalFile);
		aFile.initWithPath(src);
		if(aFile.exists() && aFile.isFile()){	// Copy the original cleopack to the tmp
			FEBE.febeCopyFile(src,cleoTmpDir.path,finalPackageName);
			tmpName = finalPackageName;
		}
	}
	var srcDir = CLEO.cleoFilesToPackageDir.path;
	var destDir = CLEO.cleoExtBUdir;
	var zipName = finalPackageName;
	var append = (parseInt(CLEO.cleoOverwriteOpt) == 2);
	FEBE.febeZip(srcDir,destDir,zipName,false,append,false);
	try{
		tmpDirDelete.remove(true);
	}catch(e){;}
	var tmp = CLEO.cleoMsg[20].replace(/%newline%/,"\n");
	tmp = tmp.replace(/%num%/,CLEO.cleoPathList.length);
	tmp += " "+ CLEO.cleoPackageName;
	CLEO.cleoAlert(tmp);
	CLEO.hideProgressBar(true);
	return true;
},

cleoGetItemsInit: function (){
	CLEO.cleoLocalize();
	// See if destination directory exists
	var prefName = "extensions.cleo.extBUdir";
	if(CLEO.cleoPrefs.prefHasUserValue(prefName)){
		CLEO.cleoExtBUdir = CLEO.cleoGetUnicharPref(prefName);
		if(CLEO.cleoExtBUdir == ""){
			FEBE.febeSetMsgs();
			var tmp = FEBE.febeMsg[3]+"\n"+FEBE.febeMsg[4].replace("FEBE","CLEO");
			CLEO.cleoAlert(tmp);
			return false;
		}	  
		var file = Cc["@mozilla.org/file/local;1"]
                  .createInstance(Ci.nsILocalFile);
		file.initWithPath(CLEO.cleoExtBUdir);
		if (!(file.exists() && file.isDirectory())){return false;}
	} else {
		FEBE.febeSetMsgs();
		var tmp = FEBE.febeMsg[3]+"\n"+FEBE.febeMsg[4].replace("FEBE","CLEO");
		CLEO.cleoAlert(tmp);
		return false;
	}
	CLEO.cleoPathList = [];
	CLEO.cleoFileList = [];
	CLEO.cleoExtName = "";
	CLEO.cleoPackageDesc = "";
	document.getElementById("cleoItems").value = CLEO.cleoMsg[25]; 
	document.getElementById("cleoPackageName").value = CLEO.cleoExtName;
	document.getElementById("cleoDescription").value = CLEO.cleoPackageDesc;
	return true;
},

cleoGetItems: function (){
	var nsIFilePicker = Ci.nsIFilePicker;
	
	var fp = Cc["@mozilla.org/filepicker;1"]
		.createInstance(nsIFilePicker);
	fp.init(window, CLEO.cleoMsg[4], nsIFilePicker.modeOpenMultiple);
	fp.appendFilter(CLEO.cleoMsg[24], "*.xpi; *.jar");
	fp.appendFilter(CLEO.cleoMsg[22], "*.xpi");
	fp.appendFilter(CLEO.cleoMsg[23], "*.jar");
	
	// Set default directory to FEBE backup directory
	if(CLEO.cleoPrefs.prefHasUserValue("extensions.febe.extBUdir")){
		CLEO.extBUdir = CLEO.cleoGetUnicharPref("extensions.febe.extBUdir");
		var dd = Cc["@mozilla.org/file/local;1"]
                     .createInstance(Ci.nsILocalFile);
		dd.initWithPath(CLEO.extBUdir);
		fp.displayDirectory = dd;
	}
	var rv = fp.show();
	if (rv == nsIFilePicker.returnOK){
		var files = fp.files;
		while(files.hasMoreElements()) {
			var file = files.getNext()
               .QueryInterface(Ci.nsILocalFile);
			CLEO.cleoPathList.push(file.path);
		}	
		document.getElementById("cleoItems").value = CLEO.cleoPathList.join("\n");
		return true;
	}
	return false;
},

cleoValidateItems: function (){
 // Make sure all selected items exist
	var tmp = document.getElementById("cleoItems").value
	var tmpArray = [];
	tmpArray = tmp.split("\n");
	CLEO.cleoPathList = [];
	CLEO.cleoFileList = [];
	var len = tmpArray.length;

	for (var i = 0; i < len; i++){
		var file = Cc["@mozilla.org/file/local;1"]
					.createInstance(Ci.nsILocalFile);
		var item = tmpArray[i];
		if(item == ""){continue;}
		if(tmpArray[i] == CLEO.cleoMsg[25]){continue;}
		var isOK = new Boolean(true);
		try{
			file.initWithPath(item);
		}catch(e){
			isOK = false;
		}
		if(isOK){isOK = CLEO.cleoTestExt(item);}
		if(isOK){isOK = (file.exists() && !file.isDirectory());}
		if (isOK == false){
			CLEO.cleoAlert(CLEO.cleoMsg[26].replace("%item%",item));
			continue;
		}
		CLEO.cleoPathList.push(file.path);
		CLEO.cleoFileList.push(file.leafName);
	}
	return true;
},

cleoTestExt: function (item){
	var len = item.length;
	var ext = item.substring(len-4);
	if(ext != ".xpi" && ext != ".jar"){return false;}
	return true;
},

cleoExtInfo: function (file){
  // Get extension info from install.rdf
	let infoObj = {};

	var rdfs = Cc["@mozilla.org/rdf/rdf-service;1"]
						.getService(Ci.nsIRDFService);
	file.append("install.rdf");
	var ioServ = Cc["@mozilla.org/network/io-service;1"]
               .getService(Ci.nsIIOService);
	var fph = ioServ.getProtocolHandler("file").QueryInterface(Ci.nsIFileProtocolHandler);
	var srcFile = fph.getURLSpecFromFile(file);
	
	var ds = rdfs.GetDataSourceBlocking(srcFile);
	var subject = rdfs.GetResource("urn:mozilla:install-manifest");
	var prefix = "http://www.mozilla.org/2004/em-rdf#";
	
  // Guid
	var predicate = rdfs.GetResource(prefix + "id");
	var id = ds.GetTarget(subject, predicate, true);
	if(id){
		infoObj.guid = id.QueryInterface(Ci.nsIRDFLiteral).Value;
	}else{
		infoObj.guid = "";
	}
	if(!infoObj.guid){infoObj.guid = CLEO.cleoMsg[27];}
	
  // Name
	var predicate = rdfs.GetResource(prefix + "name");
	var name = ds.GetTarget(subject, predicate, true);
	if(name){
		infoObj.name = name.QueryInterface(Ci.nsIRDFLiteral).Value;
	}else{
		infoObj.name = "";
	}
	if(!infoObj.name){infoObj.name = CLEO.cleoMsg[28];}

  // Version
	var predicate = rdfs.GetResource(prefix + "version");
	var version = ds.GetTarget(subject, predicate, true);
	if(version){
		infoObj.version = version.QueryInterface(Ci.nsIRDFLiteral).Value;
	}else{
		infoObj.version = "";
	}
	if(!infoObj.version){infoObj.version = CLEO.cleoMsg[29];}

  // Description
	var predicate = rdfs.GetResource(prefix + "description");
	var description = ds.GetTarget(subject, predicate, true);
	if(description){infoObj.description = description.QueryInterface(Ci.nsIRDFLiteral).Value;}
	if(!infoObj.description){infoObj.description = CLEO.cleoMsg[30];}

  // Home Page URL
	var predicate = rdfs.GetResource(prefix + "homepageURL");
	var homepageURL = ds.GetTarget(subject, predicate, true);
	if(homepageURL){infoObj.homepageURL = homepageURL.QueryInterface(Ci.nsIRDFLiteral).Value;}
	if(!infoObj.homepageURL){infoObj.homepageURL = CLEO.cleoMsg[31];}

  // Creator
	var predicate = rdfs.GetResource(prefix + "creator");
	var creator = ds.GetTarget(subject, predicate, true);
	if(creator){infoObj.creator = creator.QueryInterface(Ci.nsIRDFLiteral).Value;}
	if(!infoObj.creator){infoObj.creator = CLEO.cleoMsg[32];}

  // FX min/max version
	var firefoxID = rdfs.GetLiteral("{ec8030f7-c20a-464f-9b0e-13a3a9e97384}");
	var idArc = rdfs.GetResource(prefix + "id");
	var targetAppResource = ds.GetSource(idArc, firefoxID, true);
	if (targetAppResource){
		var minVerArc = rdfs.GetResource(prefix + "minVersion");
		var minVerVal = ds.GetTarget(targetAppResource, minVerArc, true);
		if (minVerVal){
			infoObj.fxminversion = minVerVal.QueryInterface(Ci.nsIRDFLiteral).Value;
		}else{
			infoObj.fxminversion = "??";
		}
		var maxVerArc = rdfs.GetResource(prefix + "maxVersion");
		var maxVerVal = ds.GetTarget(targetAppResource, maxVerArc, true);
		if (maxVerVal){
			infoObj.fxmaxversion = maxVerVal.QueryInterface(Ci.nsIRDFLiteral).Value;
		}else{
			infoObj.fxmaxversion = "??";
		}
	}else{
		infoObj.fxminversion = "??";
		infoObj.fxmaxversion = "??";
	}
	
  // aboutURL
	var predicate = rdfs.GetResource(prefix + "aboutURL");
	var aboutURL = ds.GetTarget(subject, predicate, true);
	if(aboutURL){infoObj.aboutURL = aboutURL.QueryInterface(Ci.nsIRDFLiteral).Value;}
	if(!infoObj.aboutURL){infoObj.aboutURL = CLEO.cleoMsg[33];}
	
  // contributors
	var predicate = rdfs.GetResource(prefix + "contributor");
	var targets = ds.GetTargets(subject, predicate, true);
	var contributors = [];
	if(targets){
		while (targets.hasMoreElements()){
			var contributor = targets.getNext();
			if (contributor instanceof Ci.nsIRDFLiteral){
				contributors.push(contributor.Value);
			}
		}
		infoObj.contributors = contributors;
	}
	return infoObj;
},

cleoValidate: function (validateOnly){
	// Check to see if everything is set and ready
	CLEO.cleoValidateItems();
	
	if(CLEO.cleoPathList.length == 0){  // No items selected
		CLEO.cleoAlert(CLEO.cleoMsg[5]);
		return false;
	}
	CLEO.cleoPackageName = document.getElementById("cleoPackageName").value
	if(CLEO.cleoPackageName.length == 0){  // No package name
		CLEO.cleoAlert(CLEO.cleoMsg[6]);
		return false;
	}
	
	if(CLEO.cleoExtBUdir == ""){
		CLEO.cleoAlert(CLEO.cleoMsg[19]);
		return false;
	}
	// See if package name already exists
	var aFile = Cc["@mozilla.org/file/local;1"]
		.createInstance(Ci.nsILocalFile);
	aFile.initWithPath(CLEO.cleoExtBUdir);
	aFile.append(CLEO.cleoPackageName + ".xpi");
	
	var validateReturn = validateOnly;
	CLEO.cleoRewriteInfoFlag = false;
	if(aFile.exists()){
		if(parseInt(CLEO.cleoOverwriteOpt) == 2){ // Append
			CLEO.cleoRewriteInfoFlag = true;
			validateReturn = true;
		}
		if(parseInt(CLEO.cleoOverwriteOpt) == 3){ // Abort
			CLEO.cleoAlert(CLEO.cleoMsg[16]);
			validateReturn = false;
		}
		if(parseInt(CLEO.cleoOverwriteOpt) == 1){ // Overwrite
			aFile.remove(false);
			validateReturn = true;
		}
	}
	
	if(validateOnly){return validateReturn;}
	CLEO.cleoDoPackage();
	return true;
},

cleoExtractCleopack: function (){
  // Extract Cleopack if it exists
	CLEO.cleoSetStage();
	var file = Cc["@mozilla.org/file/local;1"]
			.createInstance(Ci.nsILocalFile);
	
	if(CLEO.cleoInfoFileFlag == true){
		var srcName = CLEO.cleoExtName;
		var srcPath = CLEO.cleoViewPackageName;
	}else{
		var srcName = CLEO.cleoPackageName + ".xpi";
		var srcPath = CLEO.cleoMakePath(CLEO.cleoExtBUdir, srcName);
	}
		
	file.initWithPath(srcPath);
	if(!file.exists()){return true;}
	if(FEBE.febePlatform == 1){srcName = CLEO.cleoTmpName();}	// Fix for international characters
	FEBE.febeCopyFile(file.path,CLEO.cleoTmpDirName,srcName);
	
	var file = CLEO.cleoFilesToAnalizeDir.clone()
	file.createUnique(Ci.nsIFile.DIRECTORY_TYPE, 0777);
	
	CLEO.cleoFilesToAnalizeDirPath = file.path;
	var filesToAnalizeDir = file.clone();
	var sourceDir = CLEO.cleoExtBUdir;
	var destDir = CLEO.cleoFilesToAnalizeDirPath;
	var zipName = srcName;
	FEBE.febeUnzip(sourceDir,destDir,zipName);
	FEBE.febeChmod(destDir);
	
	var files = filesToAnalizeDir.directoryEntries;
	while(files.hasMoreElements()) {
		var file = files.getNext()
               .QueryInterface(Ci.nsILocalFile);              
        if(file.leafName == "CLEO.cif"){continue;}
        if(file.leafName == "install.rdf"){continue;}
        if(file.isDirectory()){continue;}
		CLEO.cleoPathList.push(file.path);
		CLEO.cleoFileList.push(file.leafName);
	}
	return true;
},

cleoExtractItem: function (cleoItem){
  // Extract an extension/theme and gather information
	CLEO.cleoSetStage();
	var aFile = Cc["@mozilla.org/file/local;1"]
				.createInstance(Ci.nsILocalFile);
	aFile.initWithPath(cleoItem);
	
	var file = CLEO.cleoTmpDir.clone()
	var cleoName = aFile.leafName;
	file.append(cleoName+".dir");
	file.createUnique(Ci.nsIFile.DIRECTORY_TYPE, 0777);
	FEBE.febeCopyFile(cleoItem,CLEO.cleoTmpDirName,cleoName)
	let cleoItemToAnalizeDirPath = file.path;
	var installRdfDir = file.clone();
	
	var sourceDir = CLEO.cleoTmpDirName;
	var destDir = cleoItemToAnalizeDirPath;
	var zipName = cleoName;
	FEBE.febeUnzip(sourceDir,destDir,zipName);
	FEBE.febeChmod(destDir);

	var thisExt = CLEO.cleoExtInfo(installRdfDir);
	var item = new Object();
	item.name = thisExt.name;
	item.version = thisExt.version;
	item.author = thisExt.creator;
	item.description = thisExt.description;
	item.fxminver = thisExt.fxminversion
	item.fxmaxver = thisExt.fxmaxversion
	item.homepage = thisExt.homepageURL;
	return item;
},

cleoCreateDescInit: function(){
	CLEO.hideProgressBar(false);
	CLEO.cleoCreateDesc();
	return true;
},

cleoCreateDesc: function (){
	if(!CLEO.cleoValidate(true)){return false;}
	if(CLEO.cleoRewriteInfoFlag){CLEO.cleoExtractCleopack()}
	CLEO.cleoDesc = [];
	let prefix = "// ";
	let tmp = prefix + CLEO.cleoMsg[11];
	tmp = tmp.replace(/%packagename%/,"Cleopack: \""+CLEO.cleoPackageName+".xpi\"");
	tmp = tmp.replace(/%date%/,CLEO.cleoLocalDate());
	CLEO.cleoDesc.push(tmp);
	tmp = prefix + CLEO.cleoMsg[38];
	CLEO.cleoVersion = CLEOstorage.getItem("cleoVersion", "???");
	tmp = tmp.replace(/%cleoversion%/,"CLEO "+CLEO.cleoVersion);
	CLEO.cleoDesc.push(tmp);
	CLEO.cleoDesc.push(prefix);
	tmp = prefix + CLEO.cleoMsg[12];
	tmp = tmp.replace(/%num%/,CLEO.cleoFileList.length);
	CLEO.cleoDesc.push(tmp);
	let tab = "     ";
	for (let i = 0; (i < CLEO.cleoPathList.length); i++){
		let thisExt = CLEO.cleoExtractItem(CLEO.cleoPathList[i]);
		CLEO.updateProgress(i+1,CLEO.cleoPathList.length,thisExt.name);
		CLEO.cleoDesc.push(prefix + CLEO.cleoFileList[i]);
		CLEO.cleoDesc.push(tab+thisExt.name);
		CLEO.cleoDesc.push(tab+CLEO.cleoMsg[34].replace(/%version%/,thisExt.version));
		CLEO.cleoDesc.push(tab+CLEO.cleoMsg[35].replace(/%author%/,thisExt.author));
		CLEO.cleoDesc.push(tab+thisExt.description);
		tmp = CLEO.cleoMsg[36];
		tmp = tmp.replace(/%minver%/,thisExt.fxminver);
		tmp = tmp.replace(/%maxver%/,thisExt.fxmaxver);
		CLEO.cleoDesc.push(tab+tmp);
		CLEO.cleoDesc.push(tab+CLEO.cleoMsg[37].replace(/%homepage%/,thisExt.homepage));
		CLEO.cleoDesc.push(tab);
	}
	document.getElementById("cleoDescription").value = CLEO.cleoDesc.join("\n");
	CLEO.hideProgressBar(true);
	return true;
},

hideProgressBar: function(bool){
	let progressBox = document.getElementById("progressBox");
	progressBox.setAttribute('hidden',bool);
	CLEO.pauseThread();
},

updateProgress: function(curVal, maxVal, fName){
	let bar = document.getElementById("cleoProgressBar");
	let txt1 = document.getElementById("progressText1");
	let txt2 = document.getElementById("progressText2");
	let txt3 = document.getElementById("progressText3");
	let pct = parseInt((curVal/(maxVal))*100);
	txt1.value = curVal+" of "+maxVal;
	bar.value = pct;
	txt2.value = pct+"%";
	txt3.value = fName;
	CLEO.pauseThread();
	return true;
},

pauseThread: function(){
	// During validation, the message "Banned or deprecated JavaScript Identifier" is generated.
	// This function is only used when updating xul progress meters.
	// As per Jorge Villalobos, this method is acceptable for this purpose.
	let currentThread = Cc['@mozilla.org/thread-manager;1'].getService(Ci.nsIThreadManager).currentThread;
	currentThread.processNextEvent(false);
	return true;
},

cleoLocalDate: function (){
  // Return a localized date/time string
	var d = new Date();
	var ds = d.toLocaleDateString();
	var ts = d.toLocaleTimeString();
	return ds+" "+ts;
},

cleoDate: function (displayPat){
/********************************************************
    *   Valid Masks:
    *   !mmmm = Long month (eg. January)
    *   !mmm = Short month (eg. Jan)
    *   !mm = Numeric date (eg. 07)
    *   !m = Numeric date (eg. 7)
    *   !dddd = Long day (eg. Monday)
    *   !ddd = Short day (eg. Mon)
    *   !dd = Numeric day (eg. 07)
    *   !d = Numeric day (eg. 7)
    *   !yyyy = Year (eg. 1999)
    *   !yy = Year (eg. 99)
   ********************************************************/
	var aDate = new Date();
    intMonth = aDate.getMonth();
    intDate = aDate.getDate();
    intDay = aDate.getDay();
    intYear = aDate.getFullYear();

    var months_long =  new Array ('January','February','March','April',
       'May','June','July','August','September','October','November','December')
    var months_short = new Array('Jan','Feb','Mar','Apr','May','Jun',
       'Jul','Aug','Sep','Oct','Nov','Dec')
    var days_long = new Array('Sunday','Monday','Tuesday','Wednesday',
       'Thursday','Friday','Saturday')
    var days_short = new Array('Sun','Mon','Tue','Wed','Thu','Fri','Sat')

    var mmmm = months_long[intMonth]
    var mmm = months_short[intMonth]
    var mm = intMonth < 9?'0'+ (1 + intMonth) + '':(1+intMonth)+'';
    var m = 1+intMonth+'';
    var dddd = days_long[intDay];
    var ddd = days_short[intDay];
    var dd = intDate<10?'0'+intDate+'':intDate+'';
    var d = intDate+'';
    var yyyy = intYear;

    century = 0;
    while((intYear-century)>=100)
        century = century + 100;

    var yy = intYear - century
    if(yy<10)
        yy = '0' + yy + '';

    displayDate = new String(displayPat);

    displayDate = displayDate.replace(/!mmmm/i,mmmm);
    displayDate = displayDate.replace(/!mmm/i,mmm);
    displayDate = displayDate.replace(/!mm/i,mm);
    displayDate = displayDate.replace(/!m/i,m);
    displayDate = displayDate.replace(/!dddd/i,dddd);
    displayDate = displayDate.replace(/!ddd/i,ddd);
    displayDate = displayDate.replace(/!dd/i,dd);
    displayDate = displayDate.replace(/!d/i,d);
    displayDate = displayDate.replace(/!yyyy/i,yyyy);
    displayDate = displayDate.replace(/!yy/i,yy);
    return displayDate;
},

cleoGetItem: function (){
	var nsIFilePicker = Ci.nsIFilePicker;
	
	var fp = Cc["@mozilla.org/filepicker;1"]
		.createInstance(nsIFilePicker);
	fp.init(window, CLEO.cleoMsg[14], nsIFilePicker.modeOpen);

	fp.appendFilter(CLEO.cleoMsg[23], "*.xpi");
	fp.appendFilters(nsIFilePicker.filterAll);
	
	// Set default directory to current backup directory
	if(CLEO.cleoPrefs.prefHasUserValue("extensions.cleo.extBUdir")){
		CLEO.cleoExtBUdir = CLEO.cleoGetUnicharPref("extensions.cleo.extBUdir");
		var dd = Cc["@mozilla.org/file/local;1"]
                     .createInstance(Ci.nsILocalFile);
		if(CLEO.cleoExtBUdir != ""){dd.initWithPath(CLEO.cleoExtBUdir);}
		fp.displayDirectory = dd;
	}
	var rv = fp.show();
	if (rv == nsIFilePicker.returnOK)
	
	CLEO.cleoPackageName = fp.file.leafName;
	CLEO.cleoViewPackageName = fp.file.path;
	
	document.getElementById("cleoViewPackageName").value = CLEO.cleoPackageName;
	return true;
},

cleoDisplayDesc: function (){
	CLEO.cleoExtName = document.getElementById("cleoViewPackageName").value;
	if(!CLEO.cleoExtName){
		CLEO.cleoAlert(CLEO.cleoMsg[13]);
		return false;
	}
	CLEO.cleoInfoFileFlag = true;
	CLEO.cleoExtractCleopack();
	CLEO.cleoInfoFileFlag = false;
	
	var file = Cc["@mozilla.org/file/local;1"]
                     .createInstance(Ci.nsILocalFile);
	file.initWithPath(CLEO.cleoFilesToAnalizeDirPath);
	file.append("CLEO.cif");
	if(!file.exists() || file.fileSize == 0){
		CLEO.cleoAlert(CLEO.cleoMsg[17]);
		return false;
	}
	
	// open an input stream from file
	var istream = Cc["@mozilla.org/network/file-input-stream;1"]
                        .createInstance(Ci.nsIFileInputStream);
	istream.init(file, 0x01, 0444, 0);
	istream.QueryInterface(Ci.nsILineInputStream);
	
	var cis = Cc["@mozilla.org/intl/converter-input-stream;1"]
						.createInstance(Ci.nsIConverterInputStream);
	
	var charset = "UTF-8";
	cis.init(istream,charset, 1024, Ci.nsIConverterInputStream.DEFAULT_REPLACEMENT_CHARACTER);
	var lis = cis.QueryInterface(Ci.nsIUnicharLineInputStream);

	// read lines into array and parse into individual fields
	FEBE.febeUDBuList = [];
	var line = {}, lines = [], hasmore;
	do {
		hasmore = lis.readLine(line);
		lines.push(line.value); 
	} while(hasmore);

	istream.close();
	cis.close();

	document.getElementById("cleoViewDescriptionTxt").value = lines.join("\n");
	return true;
},

cleoCleanUp: function (){
	// Delete temporary files and directories
	var tmpDir = Cc["@mozilla.org/file/directory_service;1"]
                 .getService(Ci.nsIProperties)
                 .get("TmpD", Ci.nsIFile);
	tmpDir.append("febe.tmp");
	if( !tmpDir.exists() || !tmpDir.isDirectory() ) {return true;}
	
	var ctmpDir = tmpDir.clone();
	ctmpDir.append("cleo.tmp");
	if( !ctmpDir.exists() || !ctmpDir.isDirectory() ) {return true;}
	ctmpDir.remove(true);
	return true;
},

cleoHideHelpIcons: function (){
	var prefName = "extensions.cleo.hidehelp";
	CLEO.cleoHideHelp = CLEO.cleoPrefs.getBoolPref(prefName);
	var c = "helpButton";
	if(CLEO.cleoHideHelp){c = "none";}
	for(var i=1; i<12; i++){
		var seq = String(i);
		if(i<10){seq = "0".concat(seq);}
		var id = "cleoHelp".concat(seq);
		try{	
			document.getElementById(id).setAttribute("class",c) ;
		}catch(e){;}
	}
	return true;
},

cleoAcceptButton: function (){
  // Show close button for Mac OS X  & Linux users because they don't get buttons for <prefwindow>
	FEBE.febeGetPlatform();
	if(FEBE.febePlatform == 2 || FEBE.febePlatform == 3){
		document.getElementById("cleoAccepteBtn").hidden = false;
	}
	return true;
},

cleoLoadHelp: function (){
	var index = window.arguments[0];
	var field = window.arguments[1];
	const HTMLNS = "http://www.w3.org/1999/xhtml";

	document.getElementById("field").value = field;
	var ch = document.getElementById('cleo-help');
	var helpTxt = ch.getString("cleo.help_"+index+".text1");
	
	// XUL pallette
	var desc = document.createElement('description');
	desc.setAttribute('flex', "1");
	var separator = document.createElement('separator');
	separator.setAttribute('class', "thin");
	var vbox = document.createElement('vbox');
	var hbox = document.createElement('hbox');
	var grid = document.createElement('grid');
	var columns = document.createElement('columns');
	var column = document.createElement('column');
	var rows = document.createElement('rows');
	var row = document.createElement('row');
	var spacer = document.createElement('spacer');
	spacer.setAttribute('flex', "1");
	var image = document.createElement('image');
	var label = document.createElement('label');
	var text = document.createElement('text');
	// HTML namespace elements
	var htmlul = document.createElementNS(HTMLNS, "html:ul");
	var htmlli = document.createElementNS(HTMLNS, "html:li");

	var helpTxtbox = document.getElementById('helptextbox');			

	switch(index){
		case "01":
			helpTxt = ch.getString("cleo.help_"+index+".text1");
			CLEO.cleoAddTxtHelpLine(helpTxtbox,helpTxt);
			helpTxtbox.appendChild(separator);
			
			var label = label.cloneNode(false);
				label.setAttribute('style', "font-weight: bold;");
				helpTxt = ch.getString("cleo.help_"+index+".text2a");
				label.setAttribute('value', helpTxt);
				helpTxtbox.appendChild(label);
			helpTxt = ch.getString("cleo.help_"+index+".text2b");
			CLEO.cleoAddTxtHelpLine(helpTxtbox,helpTxt);
			helpTxtbox.appendChild(separator);
			
			var label = label.cloneNode(false);
				label.setAttribute('style', "font-weight: bold;");
				helpTxt = ch.getString("cleo.help_"+index+".text3a");
				label.setAttribute('value', helpTxt);
				helpTxtbox.appendChild(label);
			helpTxt = ch.getString("cleo.help_"+index+".text3b");
			CLEO.cleoAddTxtHelpLine(helpTxtbox,helpTxt);
			helpTxtbox.appendChild(separator);

			var label = label.cloneNode(false);
				label.setAttribute('style', "font-weight: bold;");
				helpTxt = ch.getString("cleo.help_"+index+".text4a");
				label.setAttribute('value', helpTxt);
				helpTxtbox.appendChild(label);
			helpTxt = ch.getString("cleo.help_"+index+".text4b");
			CLEO.cleoAddTxtHelpLine(helpTxtbox,helpTxt);
			helpTxtbox.appendChild(separator);
			
			break;

		default:
			helpTxt = ch.getString("cleo.help_"+index+".text1");
			CLEO.cleoAddTxtHelpLine(helpTxtbox,helpTxt);
			break;
	}
	FEBE.febePlaySound("help");
	return true;
},

cleoAddTxtHelpLine: function (box,txt){
	var desc = document.createElement('description');
	desc.setAttribute('flex', "1");
	desc.appendChild(document.createTextNode(txt));
	box.appendChild(desc);
	var separator = document.createElement('separator');
	separator.setAttribute('class', "thin");
	box.appendChild(separator);
},

logMsg: function(msg) {
	msg = "CLEO: "+msg;
	//CLEOconsole.log(msg)
	console.log(msg)
	let cs = Cc["@mozilla.org/consoleservice;1"].getService(Ci.nsIConsoleService);
	cs.logStringMessage(msg);
	return true;
},
	
propsConsole: function(obj,title,ignoreProperties,ignoreFunctions,ignoreObjects,expandObjs,sort){
	if(title != null){
		title = "("+title+")";
	}else{
		title = "";
	}
	if (obj === null) {
		CLEO.logMsg("propsConsole called with null argument: "+title+"\n");
		return false;
	}
	if (obj === undefined) {
		CLEO.logMsg("propsConsole called with undefined argument: "+title+"\n");
		return false;
	}
	let tmp, line, pVal, type, msg, attrb = [], func = [], objects = [];
	msg = "propsConsole: "+title+"\ntypeof '"+obj+"': ("+typeof obj+"), "+((sort) ? "Sorted" : "Unsorted")+"\n";
	for(let property in obj){
		line = property;
		try{
			pVal = obj[property];
		}catch(err){
			pVal = "(Unreachable)  "+err;
		}
		try{
			type = typeof obj[property];
		}catch(e){
			type = "Unknown";
		}
		try{
			tmp = line + " = "+pVal;
		}catch(e){
			tmp = "(Unreachable)";
		}
		switch(type){
			case 'function':
				if(!ignoreFunctions) func.push(tmp);
				break;
			case 'object':
				if(ignoreObjects) continue;
				try{
					let props = JSON.stringify(obj[property]);
					if(expandObjs && props != "null") tmp += " --> "+props;
				}catch(err){;;}
				objects.push(tmp);
				break;
			default:
				if(!ignoreProperties) attrb.push(tmp);
				break;
		}
	} 
	if(func.length == 0) func.push("(none)");
	if(objects.length == 0) objects.push("(none)");
	if(attrb.length == 0) attrb.push("(none)");

	if(sort){
		attrb.sort();
		func.sort();
		objects.sort();
	}

	if(!ignoreProperties){
		msg += "\n*** PROPERTIES/ATTRIBUTES ***\n";
		for(let i in attrb){
			msg += attrb[i]+"\n";
		}
	}
	
	if(!ignoreFunctions){
		msg += "\n*** FUNCTIONS ***\n";
		for(let i in func){
			msg += func[i]+"\n";
		}
	}
	
	if(!ignoreObjects){
		msg += "\n*** OBJECTS ***\n";
		for(let i in objects){
			msg += objects[i]+"\n";
		}
	}
	CLEO.logMsg(msg);
	return true;
},

cleoOpenLink: function (URL){
	if(!URL){return false;}
	let wm = Cc["@mozilla.org/appshell/window-mediator;1"].getService(Ci.nsIWindowMediator);
	let win = wm.getMostRecentWindow("navigator:browser");
	let currBlank = false;
	try{
		let cleoGetBrowser = win.getBrowser();
		let loc = cleoGetBrowser.mCurrentTab.linkedBrowser.contentDocument.location;
		if(loc == "about:blank") currBlank = true;
		if(loc == "about:newtab") currBlank = true;
		if(cleoGetBrowser.mCurrentTab.label == "(Untitled)") currBlank = true;
	}catch(e){;;}
	if(currBlank){
		win.loadURI(URL);
	}else{
		win.openNewTabWith(URL, this.docURL, null, null);
  }
  return win;
},

}